import os
from datetime import date, datetime
from typing import Tuple

import dash
import dash_bootstrap_components as dbc
from dash import dcc, html
from dash.dependencies import Input, Output

from fieldmaster.utils import construct_source_df_from_s3, prepare_dash_graph

# this helps to find the assets folder while serving the app
assets_path = os.getcwd() + "/fieldmaster/assets"
app = dash.Dash(external_stylesheets=[dbc.themes.BOOTSTRAP], assets_folder=assets_path)

TEAMS = ["Automatoes", "AICU", "Digilog", "Grower", "IUACAAS", "TheAutomators"]

# This is an empty dict to temporaly aready trained models
CACHED_MODEL_RESULTS = {
    "Tair": {team: {} for team in TEAMS},
    "Rhair": {team: {} for team in TEAMS},
}  # type: dict

DAYS_TO_SHOW = [3, 7, 14, 30, 60, 90, 180]

CURRENT_DATE = date(2020, 3, 1)
CURRENT_TEAM = "Automatoes"


df = construct_source_df_from_s3(TEAMS)


# the style arguments for the sidebar.
SIDEBAR_STYLE = {
    "position": "fixed",
    "top": 0,
    "left": 0,
    "bottom": 0,
    "width": "20%",
    "padding": "20px 10px",
    "background-color": "#f8f9fa",
}

# the style arguments for the main content page.
CONTENT_STYLE = {"margin-left": "22%", "margin-right": "5%", "padding": "20px 10p"}
CARD_TEXT_STYLE = {"textAlign": "center", "color": "#0074D9"}

controls = dbc.FormGroup(
    [
        html.Br(),
        dcc.DatePickerSingle(
            min_date_allowed=df.index.min(),
            max_date_allowed=df.index.max(),
            initial_visible_month=date(2020, 3, 1),
            date=CURRENT_DATE,
            display_format="MMM D, YYYY",
            id="date_selection",
        ),
        html.Br(),
        html.Br(),
        dcc.Dropdown(
            options=[{"label": team, "value": team} for team in TEAMS],
            value="Automatoes",
            id="teams_dropdown",
        ),
        html.Br(),
        html.Label("Models", style={"textAlign": "left"}),
        dcc.Checklist(
            [" SARIMA", " Prophet Univariate", " Prophet Multivariate"],
            [],
            id="model_checklist",
            style={"width": "80%", "display": "inline-block"},
        ),
        html.Br(),
        html.Br(),
        html.P("Days to Show", style={"textAlign": "left"}),
        dcc.Slider(
            min=0,
            max=6,
            marks={i: str(DAYS_TO_SHOW[i]) for i in range(0, 7)},
            value=1,
            id="historical_days_slider",
        ),
        html.Br(),
    ]
)


sidebar = html.Div(
    [html.H2("Source", style={"textAlign": "center","fontsize":30}), controls],
    style=SIDEBAR_STYLE,
)

# TODO: add pass

content_first_row = dbc.Row(
    [
        dbc.Col(
            dcc.Graph(id="Tair", style={"height": "40vh"}),
        )
    ]
)

content_second_row = dbc.Row(
    [
        dbc.Col(
            dcc.Graph(id="Rhair", style={"height": "40vh"}),
        ),
    ]
)


content = html.Div(
    [
        html.Br(),
        html.H4("Temperature", style={"textAlign": "Left", "fontsize": "80px"}),
        content_first_row,
        html.Br(),
        html.H4("Relative Humidity", style={"textAlign": "Left", "fontsize": "100px"}),
        content_second_row,
    ],
    style=CONTENT_STYLE,
)

app.layout = html.Div([sidebar, content])


@app.callback(
    Output("Tair", "figure"),
    Output("model_checklist", "value"),
    [
        Input("teams_dropdown", "value"),
        Input("date_selection", "date"),
        Input("historical_days_slider", "value"),
        Input("model_checklist", "value"),
    ],
)
def update_graph_temp(
    team: str,
    start_date: str,
    days_index: int,
    models: list,
) -> Tuple[dict, list]:

    variable = dash.callback_context.outputs_list[0]["id"]
    days_to_display = DAYS_TO_SHOW[round(days_index)]

    models = update_models_to_run(models, team, start_date)

    graph, results = prepare_dash_graph(
        df,
        {
            "variable": variable,
            "team": team,
            "start_date": start_date,
            "models": models,
            "days_to_display": days_to_display,
        },
        CACHED_MODEL_RESULTS,
    )

    update_model_results(results)

    return graph, models


@app.callback(
    Output("Rhair", "figure"),
    [
        Input("teams_dropdown", "value"),
        Input("date_selection", "date"),
        Input("historical_days_slider", "value"),
        Input("model_checklist", "value"),
    ],
)
def update_graph_rh(
    team: str,
    start_date: str,
    days_index: int,
    models: list,
) -> dict:

    variable = dash.callback_context.outputs_list["id"]
    days_to_display = DAYS_TO_SHOW[round(days_index)]

    graph, results = prepare_dash_graph(
        df,
        {
            "variable": variable,
            "team": team,
            "start_date": start_date,
            "models": models,
            "days_to_display": days_to_display,
        },
        CACHED_MODEL_RESULTS,
    )

    update_model_results(results)

    return graph


def update_models_to_run(models: list, team: str, start_date: str) -> list:

    global CURRENT_TEAM
    global CURRENT_DATE

    start_date_dt = datetime.strptime(start_date, "%Y-%m-%d").date()

    if team != CURRENT_TEAM or start_date_dt != CURRENT_DATE:
        models = []

    CURRENT_TEAM = team
    CURRENT_DATE = start_date_dt

    return models


def update_model_results(results: dict) -> dict:

    CACHED_MODEL_RESULTS = results

    return CACHED_MODEL_RESULTS


if __name__ == "__main__":
    app.run_server(debug=True, port="8085")
