from __future__ import annotations

import numpy as np
import pandas as pd
from statsmodels.tsa.statespace.sarimax import SARIMAX

from fieldmaster.metrics import MAE, MAPE, RMSE


class SARIMA:
    """
    SARIMA model
    """

    def __init__(
        self,
        order: tuple,
        seasonal_order: tuple,
        confidence_level: float = 0.90,
    ) -> None:
        """
        Initialize the SARIMA model
        """

        self.order = order
        self.seasonal_order = seasonal_order
        self.confidence_level = confidence_level

    def fit(self, Xy_train: pd.DataFrame, holdout: pd.DataFrame) -> SARIMA:
        """
        Fit the SARIMA model
        """

        # Fit the SARIMA model
        model = SARIMAX(
            Xy_train.y,
            order=self.order,
            seasonal_order=self.seasonal_order,
            enforce_stationarity=False,
            enforce_invertibility=False,
        )

        # Fit the model
        self.model_ = model.fit(disp=False, maxiter=5)

        # Get the prediction for the number of days required
        forecast = self.model_.get_forecast(len(holdout))

        # Get the yhat and yhat_lower and yhat_upper
        self.yhat_ = forecast.predicted_mean
        self.confidence_ = forecast.conf_int(alpha=1 - self.confidence_level)

        self.confidence_.columns = ["yhat_upper", "yhat_lower"]

        self.metrics_ = self._calculate_all_metrics(holdout.y.values)

        return self

    def _calculate_all_metrics(self, y_true: np.ndarray) -> dict:
        """
        Calculate all metrics
        """

        y_pred = self.yhat_

        return {
            "RMSE": round(RMSE(y_true, y_pred), 3),
            "MAE": round(MAE(y_true, y_pred), 3),
            "MAPE": round(MAPE(y_true, y_pred), 3),
        }
