import random
from datetime import datetime, timedelta
from io import BytesIO
from typing import Tuple
from urllib.request import urlopen
from zipfile import ZipFile

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from fieldmaster.models.prophet_mv import PROPHETMV
from fieldmaster.models.prophet_uv import PROPHETUV
from fieldmaster.models.SARIMA import SARIMA


COLORS = [
    "rgb(31, 119, 180)",
    "rgb(255,127,14)",
    "rgb(44,160,44)",
    "rgb(214,39,40)",
    "rgb(148,103,189)",
    "rgb(140,86,75)",
    "rgb(227,119,194)",
    "rgb(127,127,127)",
    "rgb(188,189,34)",
    "rgb(23,190,207)",
]


def drop_na_columns(df: pd.DataFrame) -> pd.DataFrame:
    """
    Drop columns with more than 50% missing values
    """

    cols_na_ratio = df.isna().sum() / len(df)
    df = df.drop(cols_na_ratio[cols_na_ratio > 0.5].index, axis=1)

    return df


def drop_unique_value_columns(df: pd.DataFrame) -> pd.DataFrame:
    """
    Drop columns with only one value
    """

    unique_vals_per_col = df.nunique()
    df = df.drop(unique_vals_per_col[unique_vals_per_col == 1].index, axis=1)

    return df


def set_target_column(df: pd.DataFrame, target_column: str) -> pd.DataFrame:
    """
    Set the target column for the model
    """

    df = df.assign(y=lambda df: df[target_column])

    return df


def set_ts_column(df: pd.DataFrame) -> pd.DataFrame:
    """
    Set the ts column for the model
    This function assumes a datetime index
    """

    df = df.assign(ds=lambda df: df.index)

    return df


def select_regressors(df: pd.DataFrame, regressors: list) -> pd.DataFrame:
    """
    Returns the ts, target, and regressors columns

    """

    columns = []

    if "ds" in df.columns:
        columns.append("ds")
    if "y" in df.columns:
        columns.append("y")

    columns.extend(regressors)

    return df.loc[:, columns]


MIN_SAMPLE_POINTS_PER_DAY = 20


def sample_df(
    df: pd.DataFrame, sampling_days: int = 12, holdout_days: int = 1
) -> Tuple[pd.DataFrame, pd.DataFrame]:

    sample_size = 0

    while sample_size < MIN_SAMPLE_POINTS_PER_DAY * (sampling_days + holdout_days):

        pos = random.choice(df.index)

        Xy_train = df[
            (df.index >= pos) & (df.index < pos + timedelta(hours=24 * (sampling_days + holdout_days)))
        ].dropna()

        sample_size = len(Xy_train)

    holdout = Xy_train.tail(24)

    return Xy_train.iloc[:-24], holdout


def sample_df_from_date(
    df: pd.DataFrame, start_date_dt: datetime, historical_days: int = 8, holdout_days: int = 1
) -> Tuple[pd.DataFrame, pd.DataFrame]:

    holdout = df.loc[
        (df.index >= start_date_dt) & (df.index < (start_date_dt + timedelta(days=holdout_days)))
    ]
    Xytrain = df.loc[
        (df.index >= (start_date_dt - timedelta(days=historical_days))) & (df.index < start_date_dt)
    ]

    if len(Xytrain) < MIN_SAMPLE_POINTS_PER_DAY * (historical_days + holdout_days):
        raise ValueError(f"Not enough data to sample from {start_date_dt}")

    return Xytrain, holdout


def median_filter(
    df: pd.DataFrame, varname: str, window: int = 24, std: int = 3, plot: bool = True
) -> pd.DataFrame:
    """
    A simple median filter, removes (i.e. replace by np.nan) observations that exceed N (default = 3)
    tandard deviation from the median over window of length P (default = 24) centered around
    each observation.

    Parameters
    ----------
    df : pandas.DataFrame
        The pandas.DataFrame containing the column to filter.
    varname : string
        Column to filter in the pandas.DataFrame. No default.
    window : integer
        Size of the window around each observation for the calculation
        of the median and std. Default is 24 (time-steps).
    std : integer
        Threshold for the number of std around the median to replace
        by `np.nan`. Default is 3 (greater / less or equal).

    Returns
    -------
    dfc : pandas.Dataframe
        A copy of the pandas.DataFrame `df` with the new, filtered column `varname`
    """
    if plot:
        plt.figure(figsize=(20, 6))
        plt.plot(df.index, df[varname], label="original", color="blue", alpha=0.2)

    df = df.assign(
        window_median=lambda df: df[varname].rolling(window=window, center=True).median(),
        window_count=lambda df: df[varname].rolling(window=window, center=True, min_periods=0).count(),
        window_std=lambda df: df[varname].rolling(window=window, center=True).std(),
        outlier=lambda df: np.where(
            (
                (df[varname] <= df["window_median"] + std * df["window_std"])
                & (df[varname] >= df["window_median"] - std * df["window_std"])
            )
            | (df["window_count"] < window),
            False,
            True,
        ),
    )

    if plot:

        plt.title(f"{varname} median filter")
        plt.scatter(
            df[df.outlier].index, df.loc[df.outlier, varname], label="outliers", color="red", alpha=0.7, s=30
        )
        plt.legend()

        plt.show()

    df = df.assign(**{varname: lambda df: np.where(df.outlier, df["window_median"], df[varname])}).drop(
        ["window_median", "window_std", "window_count", "outlier"], axis=1
    )

    return df


def load_data(path: str) -> pd.DataFrame:

    return (
        load_file_from_S3(path)
        .assign(time=lambda df: pd.TimedeltaIndex(df.index, unit="d") + datetime(1900, 1, 1))
        .pipe(lambda df: median_filter(df, "Tair"))
        .pipe(drop_na_columns)
        .pipe(drop_unique_value_columns)
        .resample("H", on="time")
        .median()
        .pipe(lambda df: set_target_column(df, "Tair"))
        .pipe(set_ts_column)
    )

presigned_url = "https://tweetapplication.s3.amazonaws.com/data-20220206T105442Z-001.zip"

def load_file_from_S3(path: str) -> pd.DataFrame:

    with urlopen(presigned_url) as response:
        with ZipFile(BytesIO(response.read())) as zip_file:
            with zip_file.open(path) as file:
                return pd.read_csv(file, index_col=0, parse_dates=True)


def construct_source_df_from_s3(teams: list) -> pd.DataFrame:

    response = urlopen(presigned_url)
    zip_file = ZipFile(BytesIO(response.read()))

    df = pd.DataFrame()

    weather = (
        pd.read_csv(zip_file.open("data/Weather.csv"), index_col=0, parse_dates=True)
        .assign(time=lambda df: pd.TimedeltaIndex(df.index, unit="d") + datetime(1900, 1, 1))
        .pipe(drop_na_columns)
        .resample("H", on="time")
        .median()
    )

    for team in teams:

        greenhouse = (
            pd.read_csv(zip_file.open(f"data/{team}/GreenhouseClimate.csv"), index_col=0, parse_dates=True)
            .assign(time=lambda df: pd.TimedeltaIndex(df.index, unit="d") + datetime(1900, 1, 1))
            .pipe(drop_na_columns)
            .resample("H", on="time")
            .median()
        )

        dff = (
            greenhouse.merge(weather, how="left", left_index=True, right_index=True)
            .assign(date=lambda df: df.index.date, team=team)
            .pipe(lambda df: set_target_column(df, "Tair"))
            .pipe(drop_unique_value_columns)
            .pipe(set_ts_column)
        )

        dff["team"] = team

        df = pd.concat([df, dff], axis=0)

    return df


def load_and_sample_df(team: str, file: str) -> pd.DataFrame:

    if file == "Weather.csv":
        path = f"../datasets/{file}"
    else:
        path = f"../datasets/{team}/{file}"

    return (
        pd.read_csv(path)
        .assign(time=lambda df: pd.TimedeltaIndex(df["time"], unit="d") + datetime(1900, 1, 1))
        .pipe(drop_na_columns)
        .resample("H", on="time")
        .median()
    )


def construct_source_df(teams: list) -> pd.DataFrame:

    df = pd.DataFrame()

    for team in teams:

        dff = (
            load_and_sample_df(team, "GreenhouseClimate.csv")
            .merge(load_and_sample_df(team, "Weather.csv"), how="left", left_index=True, right_index=True)
            .assign(date=lambda df: df.index.date, team=team)
            .pipe(lambda df: set_target_column(df, "Tair"))
            .pipe(drop_unique_value_columns)
            .pipe(set_ts_column)
        )

        dff["team"] = team

        df = pd.concat([df, dff], axis=0)

    return df


NUMBER_TRAINING_DAYS = 30


# TODO: correct may 27 prediction
def prepare_dash_graph(
    df: pd.DataFrame,
    params: dict,
    model_results: dict,
) -> Tuple[dict, dict]:

    variable = params["variable"]
    team = params["team"]
    start_date = params["start_date"]
    models = params["models"]
    days_to_display = params["days_to_display"]

    model_colors = {model: COLORS[i + 1] for i, model in enumerate(models)}

    start_date_dt = datetime.strptime(start_date, "%Y-%m-%d")

    if variable == "Tair":
        df["y"] = df["Tair"]
    elif variable == "Rhair":
        df["y"] = df["Rhair"]

    dff = df[
        (df["team"] == team)
        & (df.index >= start_date_dt - timedelta(days=days_to_display))
        & (df.index <= start_date_dt)
    ]

    Xy = df[
        (df["team"] == team)
        & (df.index >= start_date_dt - timedelta(days=NUMBER_TRAINING_DAYS))
        & (df.index <= start_date_dt)
    ]

    holdout = df[
        (df["team"] == team) & (df.index >= start_date_dt) & (df.index <= start_date_dt + timedelta(days=1))
    ]

    ts_holdout = holdout.index

    for model in models:
        model_already_fitted = check_if_model_is_fitted(variable, team, start_date, model, model_results)
        if not model_already_fitted:
            fitted_results = fit_model(Xy, holdout, model)

            if start_date not in model_results[variable][team]:
                model_results[variable][team] = {start_date: fitted_results}
            else:
                model_results[variable][team][start_date][model] = fitted_results[model]

    graph = {
        "data": [],
        "layout": {"margin": {"l": 20, "r": 20, "b": 30, "t": 20}},
    }

    # TODO: ADD annotation for metrics

    graph["data"] = graph["data"] + [  # type: ignore
        {
            "x": dff.index,
            "y": dff[variable],
            "name": team,
            "mode": "lines",
            # "line": {"color": COLORS[i]},
        }
    ]

    graph["data"] = graph["data"] + [  # type: ignore
        {
            "x": ts_holdout,
            "y": holdout[variable],
            "mode": "lines",
            "showlegend": False,
            "line": {"color": "lightgray"},
        }
    ]

    for model in models:

        graph["data"] = graph["data"] + [  # type: ignore
            {
                "x": ts_holdout,
                "y": model_results[variable][team][start_date][model]["yhat"],
                "mode": "lines",
                "name": model + " forecast",
                "line": {"color": model_colors[model]},
            }
        ]

        graph["data"] = graph["data"] + [  # type: ignore
            {
                "x": ts_holdout,
                "y": model_results[variable][team][start_date][model]["confidence"]["yhat_upper"],
                "mode": "lines",
                "showlegend": False,
                "line": {"color": "rgba(255, 0, 0, 0)"},
            }
        ]

        rgb_color = model_colors[model].replace(")", ",0.1)").replace("rgb", "rgba")

        graph["data"] = graph["data"] + [  # type: ignore
            {
                "x": ts_holdout,
                "y": model_results[variable][team][start_date][model]["confidence"]["yhat_lower"],
                "mode": "lines",
                "showlegend": True,
                "name": model + " 90% ci",
                "fill": "tonexty",
                "fillcolor": rgb_color,
                "line": {"color": "rgba(255, 0, 0, 0)"},
            }
        ]

    graph["data"] = graph["data"] + [  # type: ignore
        {
            "x": [start_date_dt, start_date_dt],
            "y": [dff[variable].max(), dff[variable].max() * 1.2],
            "mode": "lines",
            "showlegend": False,
            "line": {"color": "rgba(255, 0, 0, 0)"},
        }
    ]

    graph["data"] = graph["data"] + [  # type: ignore
        {
            "x": [start_date_dt, start_date_dt],
            "y": [dff[variable].min(), dff[variable].min() * 0.8],
            "mode": "lines",
            "showlegend": False,
            "line": {"color": "rgba(255, 0, 0, 0)"},
        }
    ]

    return graph, model_results


def check_if_model_is_fitted(variable: str, team: str, date: str, model: str, model_results: dict) -> bool:
    if date in model_results[variable][team]:
        if model in model_results[variable][team][date]:
            return True

    return False


def fit_model(Xy: pd.DataFrame, holdout: pd.DataFrame, model: str) -> pd.DataFrame:

    if model == " SARIMA":
        fitted_model = SARIMA(order=(0, 1, 1), seasonal_order=(1, 1, 1, 24)).fit(Xy, holdout)

    elif model == " Prophet Univariate":
        fitted_model = PROPHETUV().fit(Xy, holdout)

    elif model == " Prophet Multivariate":
        fitted_model = PROPHETMV().fit(Xy, holdout)

    return {
        model: {
            "yhat": fitted_model.yhat_,
            "confidence": fitted_model.confidence_,
            "RMSE": fitted_model.metrics_["RMSE"],
            "MAE": fitted_model.metrics_["MAE"],
            "MAPE": fitted_model.metrics_["MAPE"],
        }
    }
